@@ -3,7 +3,7 @@
 from lxml import etree
 from Crypto.Hash import SHA256
 
-from .contants import TRASLADO, EGRESO, INGRESO, PAGO, COMPLEMENTO_CARTA_PORTE, COMPLEMENTO_PAGO
+from .contants import TRASLADO, EGRESO, INGRESO, PAGO, COMPLEMENTO_CARTA_PORTE, COMPLEMENTO_PAGOS
 from .generator_abstract import AbstractGeneratorXml
 from .utils import replace
 
@@ -110,7 +110,7 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
             dom.setAttribute('Total',str(self.data.get('total')))
             dom.appendChild(self.get_concepto_fijo())
             dom.appendChild(self.get_impuestos())
-        if self.data.get('tipo_comprobante') in [4, 5] or self.complemento in [COMPLEMENTO_PAGO, COMPLEMENTO_CARTA_PORTE]:
+        if self.data.get('tipo_comprobante') in [4, 5] or self.complemento in [COMPLEMENTO_PAGOS, COMPLEMENTO_CARTA_PORTE]:
             dom.appendChild(self.get_complementos())
         self.dom = dom
         return self.dom
