@@ -1,11 +1,9 @@
 # -*- encoding: utf-8 -*-
-import collections
-import functools
-import operator
 
 from Crypto.Hash import SHA256
 
-from .contants import TRASLADO, EGRESO, INGRESO, PAGO, COMPLEMENTO_CARTA_PORTE, COMPLEMENTO_PAGO
+from .contants import TRASLADO, EGRESO, INGRESO, PAGO, COMPLEMENTO_CARTA_PORTE, COMPLEMENTO_PAGOS, ST_RECIBO_DONATIVO, \
+    COMPLEMENTO_DONATARIAS
 from .generator_abstract import AbstractGeneratorXml
 from .utils import replace
 
@@ -15,24 +13,33 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
     def __init__(self, data):
         super(self.__class__, self).__init__(data)
         self.version = '4.0'
-        self.data = data
         self.algoritmoSha = SHA256
         self.__cadena_original = ''
-        self.impuestos_trasladados = self.data.get('impuestos_trasladados', [])
-        self.impuestos_retenidos = self.data.get('impuestos_retenidos', [])
+        if self.data:
+            self.impuestos_trasladados = self.data.get('impuestos_trasladados', [])
+            self.impuestos_retenidos = self.data.get('impuestos_retenidos', [])
+            self.impuestos_trasladados_p = self.data.get('impuestos_trasladados_p', [])
+            self.impuestos_retenidos_p = self.data.get('impuestos_retenidos_p', [])
 
     def generate_xml(self):
         self.reset_dom()
         dom = self.dom.childNodes[0]
-        if self.data.get('tipo_comprobante') == 4: #recibo donativo
+        dom.setAttribute('xmlns:cfdi', 'http://www.sat.gob.mx/cfd/4')
+        if self.data.get('tipo') == INGRESO and self.data.get('subtipo') == ST_RECIBO_DONATIVO: #recibo donativo
             dom.setAttribute('xsi:schemaLocation',
                              'http://www.sat.gob.mx/cfd/4 ' +
                              'http://www.sat.gob.mx/sitio_internet/cfd/4/cfdv40.xsd ' +
                              'http://www.sat.gob.mx/donat ' +
                              'http://www.sat.gob.mx/sitio_internet/cfd/donat/donat11.xsd')
             dom.setAttribute('xmlns:donat', 'http://www.sat.gob.mx/donat')
+        elif self.data.get('tipo') == PAGO:
+            dom.setAttribute('xmlns:pago20', 'http://www.sat.gob.mx/Pagos20')
+            dom.setAttribute('xsi:schemaLocation',
+                             'http://www.sat.gob.mx/cfd/4 ' +
+                             'http://www.sat.gob.mx/sitio_internet/cfd/4/cfdv40.xsd ' +
+                             'http://www.sat.gob.mx/Pagos20 ' +
+                             'http://www.sat.gob.mx/sitio_internet/cfd/Pagos/Pagos20.xsd')
         elif self.complemento == COMPLEMENTO_CARTA_PORTE:
-            dom.setAttribute('xmlns:cfdi', 'http://www.sat.gob.mx/cfd/4')
             dom.setAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
             dom.setAttribute('xmlns:cartaporte20', 'http://www.sat.gob.mx/CartaPorte20')
             dom.setAttribute('xsi:schemaLocation',
@@ -43,7 +50,7 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
         else:
             dom.setAttribute('xsi:schemaLocation', 'http://www.sat.gob.mx/cfd/4 http://www.sat.gob.mx/sitio_internet/cfd/4/cfdv40.xsd')
             dom.setAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
-            dom.setAttribute('xmlns:cfdi', 'http://www.sat.gob.mx/cfd/4')
+
 
         dom.setAttribute('Version', self.data.get('version', ''))
         dom.setAttribute('Serie', self.data.get('serie', ''))
@@ -52,7 +59,7 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
         dom.setAttribute('Sello', '')
         dom.setAttribute('NoCertificado', self.data.get('no_certificado'))
         dom.setAttribute('Certificado', self.data.get('certificado'))
-        dom.setAttribute('TipoDeComprobante', self.tipo_objeto)
+        dom.setAttribute('TipoDeComprobante', self.tipo_comprobante)
         dom.setAttribute('LugarExpedicion', self.data.get('lugar_expedicion', ''))
         dom.setAttribute('Exportacion', self.data.get('exportacion', '01'))
 
@@ -69,13 +76,17 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
                 raise Exception('Debe especificar el tipo de relación de los documentos')
             dom.appendChild(self.get_cfdi_relacionados(self.data.get('cfdi_relacionados', [])))
 
-        if self.tipo_objeto == INGRESO or self.tipo_objeto == EGRESO:
+        if self.tipo_comprobante == INGRESO or self.tipo_comprobante == EGRESO and not self.data.get('concepto_fijo'):
             dom.setAttribute('FormaPago', self.data.get('forma_pago'))
-            # dom.setAttribute('Descuento', '0.00') # 'Opcional'
+
+            if self.data.get('condiciones_pago'):
+                dom.setAttribute('CondicionesDePago', self.data.get('condiciones_pago'))
+
+            if self.data.get('descuento'):
+                dom.setAttribute('Descuento', self.data.get('descuento'))
+
             dom.setAttribute('MetodoPago', self.data.get('metodo_pago'))
-            # Campos que solo en ingreso no son fijos
-            dom.setAttribute('Moneda', self.data.get('moneda'))  # 'Peso Mexicano'
-            # 'Opcional solo cuando moneda es distinta de MXN'
+            dom.setAttribute('Moneda', self.data.get('moneda'))
             dom.setAttribute('TipoCambio', '1' if self.data.get('moneda') == 'MXN' else self.data.get('tipo_cambio'))
             dom.setAttribute('SubTotal', self.data.get('subtotal', ''))
             dom.setAttribute('Total', self.data.get('total', ''))
@@ -84,19 +95,33 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
                 dom.appendChild(self.get_conceptos(conceptos_list))
             if len(self.impuestos_trasladados) > 0 or len(self.impuestos_retenidos) > 0:
                 dom.appendChild(self.get_impuestos())
-        elif self.tipo_objeto == PAGO:
+        elif self.tipo_comprobante == PAGO:
             dom.setAttribute('SubTotal','0')
             dom.setAttribute('Moneda','XXX')
             dom.setAttribute('Total','0')
-            dom.setAttribute('Exportacion', self.data.get('exportacion', '01'))
             dom.appendChild(self.get_concepto_fijo())
-        elif self.tipo_objeto == TRASLADO:
+        elif self.tipo_comprobante == EGRESO and self.data.get('concepto_fijo'):
+            dom.setAttribute('FormaPago',self.data.get('forma_pago'))
+
+            if self.data.get('condiciones_pago'):
+                dom.setAttribute('CondicionesDePago', self.data.get('condiciones_pago'))
+
+            dom.setAttribute('MetodoPago', self.data.get('metodo_pago'))
+            dom.setAttribute('Moneda', self.data.get('moneda'))  # 'Peso Mexicano'
+            # 'Opcional solo cuando moneda es distinta de MXN'
+            dom.setAttribute('TipoCambio', '1' if self.data.get('moneda') == 'MXN' else self.data.get('tipo_cambio'))
+            dom.setAttribute('SubTotal', str(self.data.get('subtotal')))
+            dom.setAttribute('Total', str(self.data.get('total')))
+            dom.appendChild(self.get_concepto_fijo())
+            if bool(self.data.get('aplica_retencion', False)) or bool(self.data.get('aplica_traslado', False)):
+                dom.appendChild(self.get_impuestos())
+        elif self.tipo_comprobante == TRASLADO:
             dom.setAttribute('SubTotal','0')
             dom.setAttribute('Moneda','XXX')
             dom.setAttribute('Total','0')
             if len(conceptos_list) > 0:
                 dom.appendChild(self.get_conceptos(conceptos_list))
-        if self.data.get('tipo_comprobante') in [4, 5] or self.complemento in [COMPLEMENTO_PAGO, COMPLEMENTO_CARTA_PORTE]:
+        if self.complemento in [COMPLEMENTO_DONATARIAS, COMPLEMENTO_PAGOS, COMPLEMENTO_CARTA_PORTE]:
             dom.appendChild(self.get_complementos())
         self.dom = dom
         return self.dom
@@ -132,7 +157,7 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
             if self.data.get('receptor_residencia_fiscal') != 'MEX':
                 receptor.setAttribute('ResidenciaFiscal', self.data.get('receptor_residencia_fiscal'))
                 receptor.setAttribute('NumRegIdTrib', self.data.get('receptor_num_reg_id_trib'))
-            if self.tipo_objeto == TRASLADO and self.data.get('uso_cfdi') != 'P01':
+            if self.tipo_comprobante == TRASLADO and self.data.get('uso_cfdi') != 'P01':
                 raise Exception('El campo Uso CFDI debe contener el valor P01 - Por Definir')
             receptor.setAttribute('RegimenFiscalReceptor', self.data.get('regimen_fiscal_receptor'))
             receptor.setAttribute('UsoCFDI', self.data.get('uso_cfdi'))
@@ -147,12 +172,12 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
             concepto.setAttribute('ClaveProdServ','84111506')
             concepto.setAttribute('Cantidad','1')
             concepto.setAttribute('ClaveUnidad','ACT')
-            if self.tipo_objeto == PAGO:
+            if self.tipo_comprobante == PAGO:
                 concepto.setAttribute('Descripcion','Pago')
                 concepto.setAttribute('ValorUnitario','0')
                 concepto.setAttribute('Importe','0')
                 concepto.setAttribute('ObjetoImp', '01')
-            elif self.tipo_objeto == EGRESO:
+            elif self.tipo_comprobante == EGRESO:
                 if self.data.get('descripcion', '') != '':
                     concepto.setAttribute('Descripcion',replace(self.data.get('descripcion', '').strip()))
                 else:
@@ -170,11 +195,8 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
             if len(self.impuestos_trasladados) > 0:
                 traslados = self.dom.createElement('cfdi:Traslados')
                 total_traslados = 0
-                diferente_exento = 0
                 for traslado in self.impuestos_trasladados:
                     total_traslados += float(traslado.get('importe'))
-                    if traslado.get('tipo_factor') != 'Exento':
-                        diferente_exento += 1
                     traslados.appendChild(self.get_traslados(traslado))
                 impuestos.appendChild(traslados)
                 impuestos.setAttribute('TotalImpuestosTrasladados', str(total_traslados))
@@ -196,8 +218,10 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
         traslado.setAttribute('Base', str(traslado_data.get('base')))
         traslado.setAttribute('Impuesto', traslado_data.get('impuesto'))
         traslado.setAttribute('TipoFactor', traslado_data.get('tipo_factor'))
-        traslado.setAttribute('TasaOCuota', str(traslado_data.get('tasa_cuota')))
-        traslado.setAttribute('Importe', str(traslado_data.get('importe')))
+        if traslado_data.get('tasa_cuota'):
+            traslado.setAttribute('TasaOCuota', str(traslado_data.get('tasa_cuota')))
+        if traslado_data.get('importe'):
+            traslado.setAttribute('Importe', str(traslado_data.get('importe')))
         return traslado
 
     def get_retenciones(self, retencion_data):
@@ -209,26 +233,26 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
     def get_complementos(self):
         try:
             complemento = self.dom.createElement('cfdi:Complemento')
-            if self.data.get('tipo_comprobante') == 4:
-                complemento.appendChild(self.get_donatarias())
-            elif self.data.get('tipo_comprobante') == 5:
-                complemento.appendChild(self.get_pagos())
+            if self.complemento == COMPLEMENTO_DONATARIAS:
+                complemento.appendChild(self.get_complemento_donatarias())
+            elif self.complemento == COMPLEMENTO_PAGOS:
+                complemento.appendChild(self.get_complemento_pagos())
             elif self.complemento == COMPLEMENTO_CARTA_PORTE:
                 complemento.appendChild(self.get_complemento_cartaporte())
             return complemento
         except Exception as e:
             raise Exception("Ocurrio un error. Error::: " + str(e))
 
-    def get_donatarias(self):
+    def get_complemento_donatarias(self):
         try:
             donataria = self.dom.createElement('donat:Donatarias')
-            donataria.setAttribute('xmlns:donat','http://www.sat.gob.mx/donat')
+            donataria.setAttribute('xmlns:donat', 'http://www.sat.gob.mx/donat')
             donataria.setAttribute('xsi:schemaLocation',
-                                   'http://www.sat.gob.mx/cfd/3 ' +
-                                   'http://www.sat.gob.mx/sitio_internet/cfd/3/cfdv33.xsd ' +
+                                   'http://www.sat.gob.mx/cfd/4 ' +
+                                   'http://www.sat.gob.mx/sitio_internet/cfd/4/cfdv40.xsd ' +
                                    'http://www.sat.gob.mx/donat ' +
                                    'http://www.sat.gob.mx/sitio_internet/cfd/donat/donat11.xsd')
-            donataria.setAttribute('version','1.1')
+            donataria.setAttribute('version', '1.1')
             donataria.setAttribute('noAutorizacion', self.data.get('donataria_no_autorizacion'))
             donataria.setAttribute('fechaAutorizacion', self.data.get('donataria_fecha_autorizacion'))
             donataria.setAttribute('leyenda', self.data.get('donataria_leyenda'))
@@ -236,58 +260,141 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
         except Exception as e:
             raise Exception("Ocurrio un error. Error::: " + str(e))
 
-    def get_pagos(self):
+    def get_complemento_pagos(self):
         try:
             pagos = self.dom.createElement('pago10:Pagos')
-            pagos.setAttribute('Version','1.0')
-            pagos.setAttribute('xmlns:catCFDI','http://www.sat.gob.mx/sitio_internet/cfd/catalogos')
-            pagos.setAttribute('xmlns:catPagos','http://www.sat.gob.mx/sitio_internet/cfd/catalogos/Pagos')
-            pagos.setAttribute('xmlns:pago10','http://www.sat.gob.mx/Pagos')
-            pagos.setAttribute('xmlns:tdCFDI','http://www.sat.gob.mx/sitio_internet/cfd/tipoDatos/tdCFDI')
-            pagos.setAttribute('xmlns:xsi','http://www.w3.org/2001/XMLSchema-instance')
-            pagos.setAttribute('xsi:schemaLocation',
-                               'http://www.sat.gob.mx/Pagos http://www.sat.gob.mx/sitio_internet/cfd/Pagos/Pagos10.xsd http://www.sat.gob.mx/sitio_internet/cfd/catalogos/catCFDI.xsd http://www.sat.gob.mx/sitio_internet/cfd/tipoDatos/tdCFDI http://www.sat.gob.mx/sitio_internet/cfd/catalogos/tdCFDI.xsd http://www.sat.gob.mx/sitio_internet/cfd/catalogos/Pagos http://www.sat.gob.mx/sitio_internet/cfd/catalogos/Pagos/catPagos.xsd')
-            pago = self.dom.createElement('pago10:Pago')
-            pago.setAttribute('FechaPago', self.data.get('fecha_pago') + 'T00:00:00')
+            pagos.setAttribute('Version','2.0')
+
+            totales = self.dom.createElement('pago20:Totales')
+            totales.setAttribute('MontoTotalPagos', str(self.data.get('monto')))
+
+            for imp_ret in self.data.get('impuestos_trasladados_p'):
+                if imp_ret.get('impuesto') == '002' and imp_ret.get('tipo_factor') == 'Exento':
+                    totales.setAttribute('TotalTrasladosBaseIVAExento', str(imp_ret.get('base')))
+
+            pago = self.dom.createElement('pago20:Pago')
+            pago.setAttribute('FechaPago', self.data.get('fecha_pago'))
+
             if self.data.get('forma_pago') == '99':
                 raise Exception('La forma de pago no puede ser "Por Definir".')
+
             pago.setAttribute('FormaDePagoP', self.data.get('forma_pago'))
             pago.setAttribute('MonedaP', self.data.get('moneda_pago'))
-            if self.data.get('moneda_pago') != 'MXN':
-                pago.setAttribute('TipoCambioP', '1' if self.data.get('moneda_pago') == 'MXN' else self.data.get('tipo_cambio_pago'))
+            pago.setAttribute('TipoCambioP', self.data.get('tipo_cambio'))
             pago.setAttribute('Monto', str(self.data.get('monto')))
+
             if self.data.get('numero_operacion') != '':
-                pago.setAttribute('NumOperacion',self.data.get('numero_operacion'))
+                pago.setAttribute('NumOperacion', self.data.get('numero_operacion'))
+
             if self.data.get('bancarizado'):
                 if self.data.get('patron_cuenta_ordenante') and self.data.get('patron_cuenta_ordenante') != '':
                     if self.data.get('banco_ordenante'):
-                        pago.setAttribute('RfcEmisorCtaOrd',self.data.get('emisor_cuenta_ordenante_rfc'))
+                        pago.setAttribute('RfcEmisorCtaOrd', self.data.get('emisor_cuenta_ordenante_rfc'))
                         if self.data.get('emisor_cuenta_ordenante_rfc') == 'XEXX010101000':
-                            pago.setAttribute('NomBancoOrdExt',self.data.get('emisor_cuenta_ordenante_nombre'))
-                    pago.setAttribute('CtaOrdenante',self.data.get('emisor_cuenta_ordenante'))
+                            pago.setAttribute('NomBancoOrdExt', self.data.get('emisor_cuenta_ordenante_nombre'))
+                    pago.setAttribute('CtaOrdenante', self.data.get('emisor_cuenta_ordenante'))
                 if self.data.get('forma_pago') != '06':
                     if self.data.get('emisor_cuenta_beneficiaria_rfc'):
-                        pago.setAttribute('RfcEmisorCtaBen',self.data.get('emisor_cuenta_beneficiaria_rfc'))
-                    pago.setAttribute('CtaBeneficiario',self.data.get('emisor_cuenta_beneficiaria'))
+                        pago.setAttribute('RfcEmisorCtaBen', self.data.get('emisor_cuenta_beneficiaria_rfc'))
+                    pago.setAttribute('CtaBeneficiario', self.data.get('emisor_cuenta_beneficiaria'))
 
-            doctos = self.data.get('documentos_relacionados')
-            for cfdi in doctos:
-                docto = self.dom.createElement('pago10:DoctoRelacionado')
-                docto.setAttribute('IdDocumento', cfdi.get('uuid'))
-                docto.setAttribute('Serie', cfdi.get('serie'))
-                docto.setAttribute('Folio', cfdi.get('folio'))
-                docto.setAttribute('MonedaDR', cfdi.get('moneda'))
-                docto.setAttribute('MetodoDePagoDR', 'PPD')
-                docto.setAttribute('NumParcialidad', str(cfdi.get('num_parcialidad')))
-                docto.setAttribute('ImpSaldoAnt', str(cfdi.get('imp_saldo_ant')))
-                docto.setAttribute('ImpPagado', str(cfdi.get('imp_pagado')))
-                docto.setAttribute('ImpSaldoInsoluto', str(cfdi.get('imp_saldo_insoluto')))
+            doctos = self.data.get('doctos_relacionados')
+            for docto_relacionado in doctos:
+                docto = self.dom.createElement('pago20:DoctoRelacionado')
+                docto.setAttribute('IdDocumento', docto_relacionado.get('uuid'))
+                docto.setAttribute('Serie', docto_relacionado.get('serie'))
+                docto.setAttribute('Folio', docto_relacionado.get('folio'))
+                docto.setAttribute('MonedaDR', docto_relacionado.get('moneda'))
+                docto.setAttribute('EquivalenciaDR', docto_relacionado.get('tipo_cambio'))
+                docto.setAttribute('NumParcialidad', str(docto_relacionado.get('num_parcialidad')))
+                docto.setAttribute('ImpSaldoAnt', str(docto_relacionado.get('imp_saldo_ant')))
+                docto.setAttribute('ImpPagado', str(docto_relacionado.get('imp_pagado')))
+                docto.setAttribute('ImpSaldoInsoluto', str(docto_relacionado.get('imp_saldo_insoluto')))
+                docto.setAttribute('ObjetoImpDR', docto_relacionado.get('objeto_impuesto', '01'))
+
+                if docto_relacionado.get('objeto_impuesto', '01') != '01':
+                    docto.appendChild(self.get_impuestos_documentos_relacionados_pago(docto_relacionado))
                 pago.appendChild(docto)
+            pagos.appendChild(totales)
+            pagos.appendChild(pago)
+            if len(self.data.get('impuestos_retenidos_p')) > 0 or len(self.data.get('impuestos_trasladados_p')):
+                pago.appendChild(self.get_impuestos_pago())
             pagos.appendChild(pago)
             return pagos
         except Exception as e:
             raise Exception("Ocurrio un error. Error::: " + str(e))
 
+    def get_impuestos_documentos_relacionados_pago(self, docto_relacionado):
+        impuestos = self.dom.createElement('pago20:ImpuestosDR')
+        if len(docto_relacionado.get('retenciones', [])) > 0:
+            retenciones = self.dom.createElement('pago20:RetencionesDR')
+            for retencion in docto_relacionado.get('retenciones', []):
+                retenciones.appendChild(self.get_retenciones_documentos_relacionados(retencion))
+            impuestos.appendChild(retenciones)
+
+        if len(docto_relacionado.get('traslados', [])) > 0:
+            traslados = self.dom.createElement('pago20:TrasladosDR')
+            for traslado in docto_relacionado.get('traslados', []):
+                traslados.appendChild(self.get_traslados_documentos_relacionados(traslado))
+            impuestos.appendChild(traslados)
+        return impuestos
+
+    def get_retenciones_documentos_relacionados(self, retencion_data):
+        retencion = self.dom.createElement('pago20:RetencionDR')
+        retencion.setAttribute('BaseDR', str(retencion_data.get('base')))  # Importe del documentos_relacionados
+        retencion.setAttribute('ImpuestoDR', retencion_data.get('impuesto'))
+        retencion.setAttribute('TipoFactorDR', retencion_data.get('tipo_factor', 'Tasa'))  # Tipo factor Tasa ó Cuota
+        retencion.setAttribute('TasaOCuotaDR', str(retencion_data.get('tasa_cuota')))  # valor del impuesto que se envia
+        retencion.setAttribute('ImporteDR', str(retencion_data.get('importe')))
+        return retencion
+
+    def get_traslados_documentos_relacionados(self, traslado_data):
+        traslado = self.dom.createElement('pago20:TrasladoDR')
+        traslado.setAttribute('BaseDR', str(traslado_data.get('base')))  # Importe del documentos_relacionados
+        traslado.setAttribute('ImpuestoDR', traslado_data.get('impuesto', ''))
+        traslado.setAttribute('TipoFactorDR', traslado_data.get('tipo_factor', 'Tasa'))  # Tipo factor Tasa ó Cuota
+        if traslado_data.get('tasa_cuota'):
+            traslado.setAttribute('TasaOCuotaDR', str(traslado_data.get('tasa_cuota')))  # valor del impuesto que se envia
+        if traslado_data.get('importe'):
+            traslado.setAttribute('ImporteDR', str(traslado_data.get('importe')))
+        return traslado
+
+    def get_impuestos_pago(self):
+        try:
+            impuestos_p = self.dom.createElement('pago20:ImpuestosP')
+            if len(self.impuestos_trasladados_p) > 0:
+                traslados = self.dom.createElement('pago20:TrasladosP')
+                for traslado in self.impuestos_trasladados_p:
+                    traslados.appendChild(self.get_traslados_pago(traslado))
+                impuestos_p.appendChild(traslados)
+            if len(self.impuestos_retenidos_p) > 0:
+                retenciones = self.dom.createElement('pago20:RetencionesP')
+                total_retenciones = 0
+                for retencion in self.impuestos_retenidos_p:
+                    total_retenciones += float(retencion.get('importe', 0))
+                    retenciones.appendChild(self.get_retenciones(retencion))
+                impuestos_p.appendChild(retenciones)
+            return impuestos_p
+        except Exception as e:
+            raise Exception("Ocurrio un error. Error::: " + str(e))
+
+    def get_traslados_pago(self, traslado_data):
+        traslado = self.dom.createElement('pago20:TrasladoP')
+        traslado.setAttribute('BaseP', str(traslado_data.get('base')))
+        traslado.setAttribute('ImpuestoP', traslado_data.get('impuesto'))
+        traslado.setAttribute('TipoFactorP', traslado_data.get('tipo_factor'))
+        if traslado_data.get('tasa_cuota'):
+            traslado.setAttribute('TasaOCuotaP', str(traslado_data.get('tasa_cuota')))
+        if traslado_data.get('importe'):
+            traslado.setAttribute('ImporteP', str(traslado_data.get('importe')))
+        return traslado
+
+    def get_retenciones_pago(self, retencion_data):
+        retencion = self.dom.createElement('pago20:RetencionP')
+        retencion.setAttribute('ImpuestoP', retencion_data.get('impuesto'))
+        retencion.setAttribute('ImporteP', str(retencion_data.get('importe')))
+        return retencion
+
     def get_addenda(self):
         try:
             addenda = self.dom.createElement('cfdi:Addenda')
@@ -301,25 +408,36 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
             try:
                 cfdi_concepto = self.dom.createElement('cfdi:Concepto')
                 cfdi_concepto.setAttribute('ClaveProdServ', concepto.get('clave_prod_serv'))
+
                 if concepto.get('no_identificacion'):
                     cfdi_concepto.setAttribute('NoIdentificacion', concepto.get('no_identificacion'))
+
                 cfdi_concepto.setAttribute('Cantidad', str(concepto.get('cantidad')))
                 cfdi_concepto.setAttribute('ClaveUnidad', concepto.get('clave_unidad'))
+
                 if concepto.get('unidad'):
                     cfdi_concepto.setAttribute('Unidad', concepto.get('unidad'))
+
                 if concepto.get('descripcion', '') == '':
                     raise Exception('Se debe especificar una descripción para el concepto')
+
                 cfdi_concepto.setAttribute('Descripcion', replace(concepto.get('descripcion', '').strip()))
                 cfdi_concepto.setAttribute('ValorUnitario', str(concepto.get('valor_unitario')))
                 cfdi_concepto.setAttribute('Importe', str(concepto.get('importe')))
-                cfdi_concepto.setAttribute('ObjetoImp', concepto.get('objeto_impuesto'))  # Nuevo
-                if self.tipo_objeto != TRASLADO and concepto.get('objeto_impuesto') == '02':
+                if concepto.get('descuento'):
+                    cfdi_concepto.setAttribute('Descuento', str(concepto.get('descuento')))
+                cfdi_concepto.setAttribute('ObjetoImp', concepto.get('objeto_impuesto'))
+
+                if self.tipo_comprobante != TRASLADO and concepto.get('objeto_impuesto') == '02':
                     #revisar en el manual si lleva impuestos este tipo de comprobante
                     cfdi_concepto.appendChild(self.get_impuestos_concepto(concepto))
+
                 if concepto.get('informacion_aduanera'):
                     cfdi_concepto.appendChild(self.get_informacion_aduanera(concepto.get('informacion_aduanera')))
+
                 if concepto.get('cuenta_predial'):
                     cfdi_concepto.appendChild(self.get_cuenta_predial(concepto.get('cuenta_predial')))
+
                 conceptos.appendChild(cfdi_concepto)
             except Exception as e:
                 raise Exception("Ocurrio un error. Error::: " + str(e))
@@ -343,9 +461,9 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
         traslado = self.dom.createElement('cfdi:Traslado')
         traslado.setAttribute('Base', str(traslado_data.get('base')))  # Importe del concepto
         traslado.setAttribute('Impuesto', traslado_data.get('impuesto', ''))
-        traslado.setAttribute('TipoFactor', traslado_data.get('tipo_factor', 'Tasa'))  # Tipo factor Tasa ó Cuota
+        traslado.setAttribute('TipoFactor', traslado_data.get('tipo_factor', 'Tasa'))
         if traslado_data.get('tipo_factor') != 'Exento':
-            traslado.setAttribute('TasaOCuota', str(traslado_data.get('tasa_cuota')))  # valor del impuesto que se envia
+            traslado.setAttribute('TasaOCuota', str(traslado_data.get('tasa_cuota')))
             traslado.setAttribute('Importe', str(traslado_data.get('importe')))
         return traslado
 
@@ -412,28 +530,16 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
                 cartaporte_ubicacion = self.dom.createElement('cartaporte20:Ubicacion')
                 destination_counter = 1
                 if ubicacion_data.get('type') == 'origen':
-                    #cartaporte_ubicacion_origen = self.dom.createElement('cartaporte20:Origen')
                     cartaporte_ubicacion.setAttribute('IDUbicacion', 'OR{}'.format('000001'))
                     cartaporte_ubicacion.setAttribute('TipoUbicacion', 'Origen')
                     cartaporte_ubicacion.setAttribute('RFCRemitenteDestinatario', ubicacion_data.get('RFCRemitente'))
-                    # cartaporte_ubicacion.setAttribute('NombreRFC', ubicacion_data.get('NombreRemitente'))
                     cartaporte_ubicacion.setAttribute('FechaHoraSalidaLlegada', ubicacion_data.get('FechaHoraSalida'))
-                    #cartaporte_ubicacion_origen.setAttribute('RFCRemitente', ubicacion_data.get('RFCRemitente'))
-                    #cartaporte_ubicacion_origen.setAttribute('NombreRemitente', ubicacion_data.get('NombreRemitente'))
-                    #cartaporte_ubicacion_origen.setAttribute('FechaHoraSalida', ubicacion_data.get('FechaHoraSalida'))
-                    #cartaporte_ubicacion.appendChild(cartaporte_ubicacion_origen)
                 elif ubicacion_data.get('type') == 'destino':
                     cartaporte_ubicacion.setAttribute('IDUbicacion', 'DE00000{}'.format(destination_counter))
                     cartaporte_ubicacion.setAttribute('DistanciaRecorrida', str(ubicacion_data.get('DistanciaRecorrida')))
                     cartaporte_ubicacion.setAttribute('TipoUbicacion', 'Destino')
                     cartaporte_ubicacion.setAttribute('RFCRemitenteDestinatario', ubicacion_data.get('RFCDestinatario'))
-                    # cartaporte_ubicacion.setAttribute('NombreRFC', ubicacion_data.get('NombreDestinatario'))
                     cartaporte_ubicacion.setAttribute('FechaHoraSalidaLlegada', ubicacion_data.get('FechaHoraProgLlegada'))
-                    # cartaporte_ubicacion_destino = self.dom.createElement('cartaporte20:Destino')
-                    # cartaporte_ubicacion_destino.setAttribute('RFCDestinatario', ubicacion_data.get('RFCDestinatario'))
-                    # cartaporte_ubicacion_destino.setAttribute('NombreDestinatario', ubicacion_data.get('NombreDestinatario'))
-                    # cartaporte_ubicacion_destino.setAttribute('FechaHoraProgLlegada', ubicacion_data.get('FechaHoraProgLlegada'))
-                    #cartaporte_ubicacion.appendChild(cartaporte_ubicacion_destino)
                     total_distancia_recorrida += float(ubicacion_data.get('DistanciaRecorrida'))
                     destination_counter += 1
 
@@ -503,8 +609,6 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
                 cartaporte_tipo_transporte = self.dom.createElement('cartaporte20:Autotransporte')
                 cartaporte_tipo_transporte.setAttribute('PermSCT', transporte_data.get('PermSCT'))
                 cartaporte_tipo_transporte.setAttribute('NumPermisoSCT', transporte_data.get('NumPermisoSCT'))
-                # cartaporte_tipo_transporte.setAttribute('NombreAseg', transporte_data.get('NombreAseg'))
-                # cartaporte_tipo_transporte.setAttribute('NumPolizaSeguro', transporte_data.get('NumPolizaSeguro'))
 
                 cartaporte_identificacion_vehicular = self.dom.createElement('cartaporte20:IdentificacionVehicular')
                 cartaporte_identificacion_vehicular.setAttribute('ConfigVehicular', transporte_data.get('ConfigVehicular'))
@@ -528,44 +632,17 @@ class GeneratorXmlFourDotZero(AbstractGeneratorXml):
                         cartaporte_remolque.setAttribute('Placa', remolque_data.get('Placa'))
                         cartaporte_remolques.appendChild(cartaporte_remolque)
                     cartaporte_tipo_transporte.appendChild(cartaporte_remolques)
-
                 cartaporte_mercancias.appendChild(cartaporte_tipo_transporte)
 
             cartaporte_figura_transporte = self.dom.createElement('cartaporte20:FiguraTransporte')
-            # cartaporte_figura_transporte.setAttribute('CveTransporte', transporte_data.get('CveTransporte'))
 
             operadores = transporte_data.get('operadores', [])
-            # cartaporte_operadores = self.dom.createElement('cartaporte20:Operadores')
             for operador_data in operadores:
-                # cartaporte_operador = self.dom.createElement('cartaporte20:Operador')
-                # cartaporte_operador.setAttribute('RFCOperador', operador_data.get('RFCOperador'))
-                # cartaporte_operador.setAttribute('NumLicencia', operador_data.get('NumLicencia'))
-                # cartaporte_operador.setAttribute('NombreOperador', operador_data.get('NombreOperador'))
-
                 cartaporte_operador = self.dom.createElement('cartaporte20:TiposFigura')
                 cartaporte_operador.setAttribute('TipoFigura', '01')
                 cartaporte_operador.setAttribute('RFCFigura', operador_data.get('RFCOperador'))
                 cartaporte_operador.setAttribute('NumLicencia', operador_data.get('NumLicencia'))
 
-                # if operador_data.get('Calle', '') != '':
-                #     cartaporte_operador_domicilio = self.dom.createElement('cartaporte20:Domicilio')
-                #     cartaporte_operador_domicilio.setAttribute('Calle', operador_data.get('Calle'))
-                #     if operador_data.get('NumeroExterior', '') != '':
-                #         cartaporte_operador_domicilio.setAttribute('NumeroExterior', operador_data.get('NumeroExterior'))
-                #     if operador_data.get('Colonia', '') != '':
-                #         cartaporte_operador_domicilio.setAttribute('Colonia', operador_data.get('Colonia'))
-                #     cartaporte_operador_domicilio.setAttribute('Localidad', operador_data.get('Localidad'))
-                #     if operador_data.get('Referencia', '') != '':
-                #         cartaporte_operador_domicilio.setAttribute('Referencia', operador_data.get('Referencia'))
-                #     cartaporte_operador_domicilio.setAttribute('Municipio', operador_data.get('Municipio'))
-                #     cartaporte_operador_domicilio.setAttribute('Estado', operador_data.get('Estado'))
-                #     cartaporte_operador_domicilio.setAttribute('Pais', operador_data.get('Pais'))
-                #     cartaporte_operador_domicilio.setAttribute('CodigoPostal', operador_data.get('CodigoPostal'))
-                #     cartaporte_operador.appendChild(cartaporte_operador_domicilio)
-
-                # cartaporte_operadores.appendChild(cartaporte_operador)
-
-            # cartaporte_figura_transporte.appendChild(cartaporte_operadores)
             cartaporte_figura_transporte.appendChild(cartaporte_operador)
             cartaporte.appendChild(cartaporte_figura_transporte)
 
